const ObjectTypeMap = {
    '[object String]': 'string',
    '[object Number]': 'number',
    '[object Boolean]': 'boolean',
    '[object Symbol]': 'symbol',
    '[object Undefined]': 'undefined',
    '[object Null]': 'null',
    '[object Function]': 'function',
    '[object Date]': 'date',
    '[object Array]': 'array',
    '[object RegExp]': 'regexp',
    '[object Error]': 'error',
    '[object Object]': 'object'
};
/**
 * 获取输入参数的类型
 * @param input
 * @returns {string}
 */
Object.type = (input) => {
    const str = Object.prototype.toString.call(input);
    return ObjectTypeMap[str] || str;
};
//
Object.isObject = (input) => {
    return Object.type(input) === 'object';
};

Object.isString = (input) => {
    return Object.type(input) === 'string';
};

Object.isNumber = (input) => {
    return Object.type(input) === 'number';
};

Object.isBoolean = (input) => {
    return Object.type(input) === 'boolean';
};

Object.isSymbol = (input) => {
    return Object.type(input) === 'symbol';
};

Object.isUndefined = (input) => {
    return Object.type(input) === 'undefined';
};

Object.isNull = (input) => {
    return Object.type(input) === 'null';
};

Object.isFunction = (input) => {
    return Object.type(input) === 'function';
};

Object.isDate = (input) => {
    return Object.type(input) === 'date';
};

Object.isArray = (input) => {
    return Object.type(input) === 'array';
};

Object.isRegExp = (input) => {
    return Object.type(input) === 'regexp';
};

Object.isError = (input) => {
    return Object.type(input) === 'error';
};

/**
 * 对象stringify时默认的属性转换器
 * @param value
 * @param key
 * @param object
 * @returns {*}
 * @constructor
 */
const OBJECT_STRINGIFY_DEFAULT_REPLACER = (value, key, object) => {
    if (value == null) {
        return;
    }
    return `${key}:${value}`;
};

/**
 * 对象stringify
 * @param object    转换的对象
 * @param replacer
 * @param joiner
 * @returns {string|null}
 */
Object.stringify = (object, {
    replacer,
    joiner = ';'
} = {}) => {
    if (object == null || !Object.isObject(object)) {
        return null;
    }
    const keyValues = [];
    let propReplacer = Object.isFunction(replacer) ? replacer : OBJECT_STRINGIFY_DEFAULT_REPLACER;
    Object.keys(object).forEach(key => {
        const value = object[key];
        const output = propReplacer(value, key, object);
        if (output == null) {
            return;
        }
        keyValues.push(output);
    });
    return keyValues.join(joiner);
};
//
String.informatUuid = function () {
    return informatUuid();
};
String.uuid = function (len) {
    return nanoUuid(len);
};
//
String.prototype.format = function () {
    var args = arguments;
    return this.replace(/{([0-9]+)}/g, function (match, index) {
        return typeof args[index] == 'undefined' ? match : args[index];
    });
};

String.isMobile = function (s) {
    if (String.isEmpty(s)) {
        return true;
    }
    return /^1\d{10}$/.test(s);
};

String.isEmail = function (s) {
    if (String.isEmpty(s)) {
        return true;
    }
    return /^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/.test(s);
};

String.isUrl = function (s) {
    if (String.isEmpty(s)) {
        return false;
    }
    return /^(https?):\/\//i.test(s);
};

String.isNotEmpty = function (s) {
    if (s == null) {
        return false;
    }
    if (`${s}`.trim().length === 0) {
        return false;
    }
    return true;
};

String.isEmpty = function (s) {
    if (s == null) {
        return true;
    }
    if (!Object.isString(s)) {
        return false;
    }
    if (`${s}`.trim().length === 0) {
        return true;
    }
    return false;
};
//
String.prototype.convertCase = function (delimiter = '_') {
    return this.match(/[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+/g)
        .map(x => x.toLowerCase())
        .join(delimiter);
};
// 首字母小写
String.prototype.firstCharToLowCase = function () {
    return this.charAt(0).toLowerCase() + this.slice(1);
};
// 首字母大写
String.prototype.firstCharToUpperCase = function () {
    return this.charAt(0).toUpperCase() + this.slice(1);
};
// 驼峰转
String.prototype.toKebabCase = function () {
    return this.replace(/[A-Z]/g, letter => `-${letter.toLowerCase()}`);
};
//
String.prototype.htmlEscape = function () {
    return this.replace(/&/g, '&amp;')
        .replace(/"/g, '&quot;')
        .replace(/'/g, '&#39;')
        .replace(/</g, '&lt;')
        .replace(/>/g, '&gt;');
};

Array.isNotEmpty = (array) => {
    if (!array) {
        return false;
    }
    if (!Array.isArray(array)) {
        return false;
    }
    return array.length > 0;
};

Array.isEmpty = (array) => {
    if (!array) {
        return true;
    }
    if (!Array.isArray(array)) {
        return true;
    }
    return array.length === 0;
};
Array.prototype.size = function () {
    return this.length;
};
//
Array.prototype.remove = function (item) {
    const index = this.indexOf(item);
    if (index === -1) {
        return;
    }
    this.splice(index, 1);
};
Array.prototype.removeById = function (id) {
    this.removeByKey('id', id);
};
Array.prototype.removeByKey = function (key, value) {
    const index = this.findIndex(item => item[key] == value);
    if (index === -1) {
        return;
    }
    this.splice(index, 1);
};

/**
 * 数组添加元素，元素已存在则跳过
 * @param item
 */
Array.prototype.uniqueAdd = function (item) {
    if (this.indexOf(item) === -1) {
        this.push(item);
    }
    return this;
};

/**
 * 数组合并，array中的元素已存在则跳过
 * @param array
 */
Array.prototype.uniqueConcat = function (array) {
    if (!Array.isArray(array)) {
        return this;
    }
    array.forEach(item => {
        if (this.indexOf(item) === -1) {
            this.push(item);
        }
    });
    return this;
};
Array.prototype.toggle = function (item) {
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        if (t == item) {
            this.splice(i, 1);
            return;
        }
    }
    this.push(item);
};
Array.prototype.toggleObjectById = function (item) {
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        if (t.id == item.id) {
            this.splice(i, 1);
            return;
        }
    }
    this.push(item);
};
Array.prototype.findObject = function (item) {
    return this.find(t => t == item);
};
Array.prototype.findById = function (id) {
    return this.findByKey('id', id);
};
Array.prototype.findIndexById = function (id) {
    return this.findIndexByKey('id', id);
};
Array.prototype.findIndexByKey = function (key, value) {
    for (var i = 0; i < this.length; i++) {
        if (this[i][key] == value) {
            return i;
        }
    }
    return -1;
};
Array.prototype.findByKey = function (key, value) {
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        if (t[key] == value) {
            return t;
        }
    }
    return null;
};

Array.prototype.match = function (array) {
    if (array.length === 0) {
        return true;
    }
    const arrayObject = {};
    array.forEach(t => {
        arrayObject[t] = true;
    });
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        if (arrayObject[t]) {
            return true;
        }
    }
    return false;
};

Array.prototype.toObject = function () {
    const result = {};
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        result[t.id] = t;
    }
    return result;
};

Array.prototype.toObjectWithKey = function (key = 'id') {
    const result = {};
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        result[t[key]] = t;
    }
    return result;
};

Array.prototype.cloneProp = function (propArray) {
    const result = [];
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        const item = {};
        propArray.forEach(p => {
            item[p] = t[p];
        });
        result.push(item);
    }
    return result;
};

Array.prototype.distinct = function () {
    const retMap = {};
    const result = [];
    this.forEach(item => {
        if (retMap[item] == null) {
            result.push(item);
            retMap[item] = true;
        }
    });
    return result;
};

Array.prototype.propList = function (prop) {
    const result = [];
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        result.push(t[prop]);
    }
    return result;
};

Array.prototype.walk = function (callback, parent) {
    for (let i = 0; i < this.length; i++) {
        const t = this[i];
        callback(t, parent);
        if (t.children) {
            t.children.walk(callback, t);
        }
    }
};
Array.prototype.treeRemove = function (t) {
    for (let i = 0; i < this.length; i++) {
        if (this[i] == t) {
            this.splice(i, 1);
            return;
        }
    }
    this.forEach(item => {
        if (item.children) {
            item.children.treeRemove(t);
        }
    });
};
Array.prototype.treeRemoveByKey = function (value, key = 'id') {
    for (let i = 0; i < this.length; i++) {
        if (this[i][key] == value) {
            this.splice(i, 1);
            return;
        }
    }
    this.forEach(item => {
        if (item.children) {
            item.children.treeRemoveByKey(value, key);
        }
    });
};
Array.prototype.treeRemoveById = function (id) {
    return this.treeRemoveByKey(id, 'id');
};

Array.prototype.treeFindByKey = function (value, key = 'id') {
    for (let i = 0; i < this.length; i++) {
        if (this[i][key] == value) {
            return this[i];
        }
        if (this[i].children) {
            const rs = this[i].children.treeFindByKey(value, key);
            if (rs) return rs;
        }
    }
};
Array.prototype.treeFindById = function (id) {
    return this.treeFindByKey(id, 'id');
};

/**
 * 剪枝函数，影响原有树结构
 * @param preservedLeaves 保留的节点
 * @param key 剪枝对象的key
 */
Array.prototype.treePruneByKey = function (preservedLeaves, key = 'id') {
    for (let i = 0; i < this.length;) {
        const node = this[i];
        if (node.children && node.children.length > 0) {
            node.children.treePruneByKey(preservedLeaves, key);
        }
        if ((!node.children || node.children.length === 0) && !preservedLeaves.includes(node[key])) {
            this.splice(i, 1);
        } else {
            i++;
        }
    }
};

/**
 * 层级结构数据中过滤数据
 * @param callback
 * @param idKey
 */
Array.prototype.walkFilter = function (callback, idKey = 'id') {
    const list = [];
    this.walk((item, parent) => {
        list.push({
            ...item,
            parentId: parent ? parent[idKey] : null
        });
    });
    //
    const newList = list.filter(item => {
        return callback(item);
    });
    return newList.asTree();
};

/**
 * 层级结构数据中过滤数据
 * @param accept
 */
Array.prototype.treeFilter = function (accept) {
    const result = [];
    this.forEach(item => {
        if (accept(item)) {
            result.push(item);
            if (item.children) {
                item.children = item.children.treeFilter(accept);
            }
        }
    });
    return result;
};

/**
 * 展开树，项是浅clone出来的对象
 * @param idKey
 * @returns {*[]}
 */
Array.prototype.flatTree = function (idKey = 'id') {
    const list = [];
    this.walk((item, parent) => {
        list.push({
            ...item,
            parentId: parent ? parent[idKey] : null
        });
    });
    return list;
};

/**
 * 展开树，项是对象引用
 * @param idKey
 * @returns {*[]}
 */
Array.prototype.originFlatTree = function (idKey = 'id') {
    const list = [];
    this.walk((item, parent) => {
        list.push(Object.assign(item, {
            parentId: parent ? parent[idKey] : null
        }));
    });
    return list;
};

/**
 * 项是对象引用
 * @returns {*[]}
 */
Array.prototype.asList = function () {
    const result = [];
    this.walk(t => {
        if (!t.isDirectory) {
            result.push(t);
        }
    });
    return result;
};

//
Array.prototype.asTree = function () {
    const allMap = {};
    for (let i = 0; i < this.length; i++) {
        const item = this[i];
        item.children = [];
        allMap[item.id] = item;
    }
    //
    const result = [];
    for (let j = 0; j < this.length; j++) {
        const jitem = this[j];
        if (jitem.parentId == 0 || jitem.parentId == null || jitem.parentId == undefined || jitem.parentId == '/') {
            result.push(jitem);
        } else {
            const parent = allMap[jitem.parentId];
            if (parent) {
                parent.children.push(jitem);
            } else {
                result.push(jitem);
            }
        }
    }
    return result;
};
Array.prototype.addSet = function (o) {
    if (this.indexOf(o) > -1) {
        return;
    }
    this.push(o);
};

Array.prototype.addSetWithList = function (list) {
    if (Array.isEmpty(list)) {
        return;
    }
    list.forEach(item => {
        this.addSetWithId(item);
    });
};

Array.prototype.addSetWithId = function (o) {
    this.addSetWithKey(o, 'id');
};

Array.prototype.addSetWithKey = function (o, key = 'id') {
    if (o == null || key == null) {
        if (this.indexOf(o) > -1) {
            return;
        }
        this.push(o);
        return;
    }
    const index = this.findIndexByKey(key, o[key]);
    if (index > -1) {
        return;
    }
    this.push(o);
};
